#include <math.h>

#include "bs_blob.h"
#include "bs_mem.h"
#include "bs_assert.h"
#include "bs_util.h"
#include "bs_vector.h"

#define BS_BLOB_NUM_POINT_MASSES 4
#define BS_BLOB_NUM_JOINTS 6

bs_blob_t* 
bs_blob_create(bs_float_t xPos, 
               bs_float_t yPos, 
               bs_float_t radius)
{
  bs_blob_t *pBlob; 
  bs_float_t shortConst, longConst; 
  
  BS_MEM_ALLOC_TYPE(pBlob, bs_blob_t, 1);   

  pBlob->pMiddle = bs_pointmass_create(xPos, yPos, 1.0); 
  pBlob->radius = radius; 
  pBlob->faceStyle = BS_BLOB_FACE_STYLE_MOUTH_CLOSED; 
  pBlob->eyeStyle = BS_BLOB_EYE_STYLE_OPEN; 

  
  BS_MEM_ALLOC_TYPE(pBlob->pPointmasses, bs_pointmass_t*, BS_BLOB_NUM_POINT_MASSES); 
  pBlob->pPointmasses[0] = bs_pointmass_create(xPos - 0.7f * radius, yPos - 0.7f * radius, 1.0f); 
  pBlob->pPointmasses[1] = bs_pointmass_create(xPos + 0.7f * radius, yPos - 0.7f * radius, 1.0f); 
  pBlob->pPointmasses[2] = bs_pointmass_create(xPos + 0.7f * radius, yPos + 0.7f * radius, 1.0f); 
  pBlob->pPointmasses[3] = bs_pointmass_create(xPos - 0.7f * radius, yPos + 0.7f * radius, 1.0f); 
  
  
  BS_MEM_ALLOC_TYPE(pBlob->pSticks, bs_stick_t*, BS_BLOB_NUM_POINT_MASSES); 
  pBlob->pSticks[0] = bs_stick_create(pBlob->pPointmasses[0], pBlob->pPointmasses[1]); 
  pBlob->pSticks[1] = bs_stick_create(pBlob->pPointmasses[1], pBlob->pPointmasses[2]); 
  pBlob->pSticks[2] = bs_stick_create(pBlob->pPointmasses[2], pBlob->pPointmasses[3]); 
  pBlob->pSticks[3] = bs_stick_create(pBlob->pPointmasses[3], pBlob->pPointmasses[0]); 
  

  shortConst = 0.99f; 
  longConst = 1.05f; 
  BS_MEM_ALLOC_TYPE(pBlob->pJoints, bs_joint_t*, BS_BLOB_NUM_JOINTS); 
  pBlob->pJoints[0] = bs_joint_create(pBlob->pPointmasses[0], pBlob->pMiddle, shortConst, longConst); 
  pBlob->pJoints[1] = bs_joint_create(pBlob->pPointmasses[1], pBlob->pMiddle, shortConst, longConst); 
  pBlob->pJoints[2] = bs_joint_create(pBlob->pPointmasses[2], pBlob->pMiddle, shortConst, longConst); 
  pBlob->pJoints[3] = bs_joint_create(pBlob->pPointmasses[3], pBlob->pMiddle, shortConst, longConst); 
  
  pBlob->pJoints[4] = bs_joint_create(pBlob->pPointmasses[0], pBlob->pPointmasses[2], shortConst, longConst); 
  pBlob->pJoints[5] = bs_joint_create(pBlob->pPointmasses[1], pBlob->pPointmasses[3], shortConst, longConst); 

  return pBlob; 
}

void 
bs_blob_destroy(bs_blob_t *pBlob)
{
  int i; 
  
  BS_ASSERT(pBlob != NULL); 

  bs_pointmass_destroy(pBlob->pMiddle); 
  
  BS_ASSERT(pBlob->pPointmasses != NULL);   
  for(i = 0; i < BS_BLOB_NUM_POINT_MASSES; i++)
  {
    bs_pointmass_destroy(pBlob->pPointmasses[i]); 
  }
  BS_MEM_FREE(pBlob->pPointmasses); 
  
  BS_ASSERT(pBlob->pSticks != NULL); 
  for(i = 0; i < BS_BLOB_NUM_POINT_MASSES; i++)
  {
    bs_stick_destroy(pBlob->pSticks[i]); 
  }
  BS_MEM_FREE(pBlob->pSticks); 
  
  BS_ASSERT(pBlob->pJoints != NULL); 
  for(i = 0; i < BS_BLOB_NUM_JOINTS; i++)
  {
    bs_joint_destroy(pBlob->pJoints[i]); 
  }
  BS_MEM_FREE(pBlob->pJoints); 
  
  BS_MEM_FREE(pBlob); 
}

bs_float_t 
bs_blob_get_radius(bs_blob_t *pBlob)
{
  BS_ASSERT(pBlob != NULL); 
  return pBlob->radius;   
}

void
bs_blob_scale(bs_blob_t *pBlob, 
              bs_float_t scaleFactor)
{
  int i; 

  BS_ASSERT(pBlob != NULL); 
  BS_ASSERT(scaleFactor > 0.0f); 

  for(i = 0; i < BS_BLOB_NUM_JOINTS; i++)
  {
    bs_joint_scale(pBlob->pJoints[i], scaleFactor); 
  }    
  for(i = 0; i < BS_BLOB_NUM_POINT_MASSES; i++)
  {
    bs_stick_scale(pBlob->pSticks[i], scaleFactor); 
  }
  pBlob->radius *= scaleFactor;
}

bs_pointmass_t* 
bs_blob_get_middle(bs_blob_t *pBlob)
{
  BS_ASSERT(pBlob != NULL);
  return pBlob->pMiddle;   
}

bs_float_t 
bs_blob_get_x_pos(bs_blob_t *pBlob)
{
  BS_ASSERT(pBlob != NULL);
  return bs_pointmass_get_x_pos(pBlob->pMiddle);   
}

bs_float_t 
bs_blob_get_y_pos(bs_blob_t *pBlob)
{
  BS_ASSERT(pBlob != NULL);
  return bs_pointmass_get_y_pos(pBlob->pMiddle);   
}

void 
bs_blob_move(bs_blob_t *pBlob, 
             bs_float_t dt)
{
  int i; 

  BS_ASSERT(pBlob != NULL); 
  BS_ASSERT(pBlob->pPointmasses != NULL); 
  BS_ASSERT(pBlob->pMiddle != NULL); 
  
  for(i = 0; i < BS_BLOB_NUM_POINT_MASSES; i++)
  {
    bs_pointmass_move(pBlob->pPointmasses[i], dt); 
  }    
  bs_pointmass_move(pBlob->pMiddle, dt); 
}

void 
bs_blob_sc(bs_blob_t *pBlob, 
           bs_env_t *pEnv)
{
  int i, j;

  BS_ASSERT(pBlob != NULL); 
  BS_ASSERT(pBlob->pPointmasses != NULL); 
  BS_ASSERT(pBlob->pSticks != NULL); 
  BS_ASSERT(pBlob->pJoints != NULL); 

  for(j = 0; j < 20; j++)
  {
    for(i = 0; i < BS_BLOB_NUM_POINT_MASSES; i++)
    {
      if(bs_env_collision(pEnv, bs_pointmass_get_pos(pBlob->pPointmasses[i]), 
         bs_pointmass_get_prev_pos(pBlob->pPointmasses[i])) == TRUE)
      {
        bs_pointmass_set_friction(pBlob->pPointmasses[i], 0.75f); 
      }
      else 
      {
        bs_pointmass_set_friction(pBlob->pPointmasses[i], 0.01f);         
      }
    }  
    for(i = 0; i < BS_BLOB_NUM_JOINTS; i++)
    {
      bs_joint_sc(pBlob->pJoints[i]); 
    }
    for(i = 0; i < BS_BLOB_NUM_POINT_MASSES; i++)
    {
      bs_stick_sc(pBlob->pSticks[i]); 
    }  
  }
}

void 
bs_blob_set_force(bs_blob_t *pBlob, 
                  bs_vector_t *pForce)
{
  int i; 
  
  BS_ASSERT(pBlob != NULL); 
  BS_ASSERT(pBlob->pPointmasses != NULL); 
  BS_ASSERT(pBlob->pMiddle != NULL);  
  
  for(i = 0; i < BS_BLOB_NUM_POINT_MASSES; i++)
  {
    bs_pointmass_set_force(pBlob->pPointmasses[i], pForce);  
  }    
  bs_pointmass_set_force(pBlob->pMiddle, pForce);  
}

void 
bs_blob_add_force(bs_blob_t *pBlob, 
                  bs_vector_t *pForce)
{
  int i; 

  BS_ASSERT(pBlob != NULL); 
  BS_ASSERT(pBlob->pPointmasses != NULL); 
  BS_ASSERT(pBlob->pMiddle != NULL);  
  
  for(i = 0; i < BS_BLOB_NUM_POINT_MASSES; i++)
  {
    bs_pointmass_add_force(pBlob->pPointmasses[i], pForce);  
  }    
  bs_pointmass_add_force(pBlob->pMiddle, pForce);  
}

static void 
bs_blob_randomize_face_style(bs_blob_t *pBlob)
{
  BS_ASSERT(pBlob != NULL);

  if(pBlob->faceStyle == BS_BLOB_FACE_STYLE_MOUTH_CLOSED && bs_util_random_float() < 0.05f)
  {
    pBlob->faceStyle = BS_BLOB_FACE_STYLE_MOUTH_OPEN; 
  }
  else if(pBlob->faceStyle == BS_BLOB_FACE_STYLE_MOUTH_OPEN && bs_util_random_float() < 0.1f)
  {
    pBlob->faceStyle = BS_BLOB_FACE_STYLE_MOUTH_CLOSED; 
  }
  
  if(pBlob->eyeStyle == BS_BLOB_EYE_STYLE_OPEN && bs_util_random_float() < 0.025f)
  {
    pBlob->eyeStyle = BS_BLOB_EYE_STYLE_BLINK; 
  }
  else if(pBlob->eyeStyle == BS_BLOB_EYE_STYLE_BLINK && bs_util_random_float() < 0.3f)
  {
    pBlob->eyeStyle = BS_BLOB_EYE_STYLE_OPEN; 
  }
}

static void 
bs_blob_draw_mouth_closed(bs_blob_t *pBlob, 
                          cairo_t *pCairo, 
                          bs_float_t rsf)
{
  BS_ASSERT(pBlob != NULL);
  BS_ASSERT(pCairo != NULL);

  cairo_new_path(pCairo); 
  cairo_set_line_width(pCairo, 2.0 / rsf);   
  cairo_arc(pCairo, 0.0, 0.0, 0.25f, M_PI, FALSE); 
  cairo_stroke(pCairo); 
}

static void 
bs_blob_draw_mouth_open(bs_blob_t *pBlob, 
                        cairo_t *pCairo, 
                        bs_float_t rsf)
{
  BS_ASSERT(pBlob != NULL);
  BS_ASSERT(pCairo != NULL);

  cairo_new_path(pCairo); 
  cairo_set_line_width(pCairo, 2.0 / rsf);   
  cairo_arc(pCairo, 0.0, 0.0, 0.25f, M_PI, FALSE); 
  cairo_close_path(pCairo); 
  cairo_fill(pCairo); 
}

static void 
bs_blob_draw_eyes_open(bs_blob_t *pBlob, 
                       cairo_t *pCairo, 
                       bs_float_t rsf)
{
  BS_ASSERT(pBlob != NULL);
  BS_ASSERT(pCairo != NULL);

  cairo_set_line_width(pCairo, 1.0 / rsf);
  
  cairo_new_path(pCairo); 
  cairo_arc(pCairo, -0.15, 0.20, 0.12, 0.0, 2.0 * M_PI); 
  cairo_close_path(pCairo); 
  cairo_stroke(pCairo); 
    
  cairo_new_path(pCairo); 
  cairo_arc(pCairo, 0.15, 0.20, 0.12, 0.0, 2.0 * M_PI); 
  cairo_close_path(pCairo); 
  cairo_stroke(pCairo); 
  
  cairo_new_path(pCairo);
  cairo_arc(pCairo, -0.15, 0.17, 0.06, 0.0, 2.0 * M_PI); 
  cairo_close_path(pCairo);
  cairo_fill(pCairo);   

  cairo_new_path(pCairo);
  cairo_arc(pCairo, 0.15, 0.17, 0.06, 0.0, 2.0 * M_PI); 
  cairo_close_path(pCairo);
  cairo_fill(pCairo);   
}

static void 
bs_blob_draw_eyes_blink(bs_blob_t *pBlob, 
                        cairo_t *pCairo, 
                        bs_float_t rsf)
{
  BS_ASSERT(pBlob != NULL);
  BS_ASSERT(pCairo != NULL);

  cairo_set_line_width(pCairo, 1.0 / rsf);

  cairo_new_path(pCairo); 
  cairo_arc(pCairo, -0.15, 0.20, 0.12, 0.0, 2.0 * M_PI); 
  cairo_close_path(pCairo); 
  cairo_stroke(pCairo); 
    
  cairo_new_path(pCairo); 
  cairo_arc(pCairo, 0.15, 0.20, 0.12, 0.0, 2.0 * M_PI); 
  cairo_close_path(pCairo); 
  cairo_stroke(pCairo); 

  cairo_new_path(pCairo); 
  cairo_move_to(pCairo, -0.25, 0.20);
  cairo_line_to(pCairo, -0.05, 0.20); 
  cairo_stroke(pCairo);  

  cairo_new_path(pCairo); 
  cairo_move_to(pCairo,  0.25, 0.20);
  cairo_line_to(pCairo,  0.05, 0.20); 
  cairo_stroke(pCairo);   
}

static void 
bs_blob_draw_ooh_face(bs_blob_t *pBlob, 
                      cairo_t *pCairo, 
                      bs_float_t rsf)
{
  BS_ASSERT(pBlob != NULL);
  BS_ASSERT(pCairo != NULL);
  
  bs_blob_draw_mouth_open(pBlob, pCairo, rsf); 
  
  cairo_set_line_width(pCairo, 2.0 / rsf); 
  
  cairo_new_path(pCairo); 
  cairo_move_to(pCairo, -0.25, 0.3); 
  cairo_line_to(pCairo, -0.05, 0.2); 
  cairo_line_to(pCairo, -0.25, 0.1); 
  cairo_stroke(pCairo); 

  cairo_new_path(pCairo); 
  cairo_move_to(pCairo,  0.25, 0.3); 
  cairo_line_to(pCairo,  0.05, 0.2); 
  cairo_line_to(pCairo,  0.25, 0.1); 
  cairo_stroke(pCairo); 
}

static void 
bs_blob_draw_face(bs_blob_t *pBlob, 
                  cairo_t *pCairo, 
                  bs_float_t scaleFactor)
{
  bs_float_t xOff, yOff; 
  bs_vector_t up, orientation; 
  bs_float_t angle; 
  bs_float_t rsf; 

  BS_ASSERT(pBlob != NULL);
  BS_ASSERT(pCairo != NULL);

  rsf = pBlob->radius * scaleFactor; 
  
  cairo_save(pCairo); 
  cairo_set_source_rgb(pCairo, 0.0, 0.0, 0.0); 
  
  BS_VECTOR_SET(&orientation, bs_pointmass_get_pos(pBlob->pPointmasses[0])); 
  BS_VECTOR_ADD(&orientation, bs_pointmass_get_pos(pBlob->pPointmasses[1])); 
  BS_VECTOR_SUB(&orientation, bs_pointmass_get_pos(pBlob->pPointmasses[2])); 
  BS_VECTOR_SUB(&orientation, bs_pointmass_get_pos(pBlob->pPointmasses[3])); 
  
  xOff = (bs_pointmass_get_x_pos(pBlob->pMiddle) + BS_VECTOR_GET_X(&orientation) * 0.2f) * scaleFactor; 
  yOff = (bs_pointmass_get_y_pos(pBlob->pMiddle) + BS_VECTOR_GET_Y(&orientation) * 0.2f) * scaleFactor;
  cairo_translate(pCairo, xOff, yOff); 
  
  bs_vector_init(&up, 0.0f, 1.0f); 
  angle = acos(bs_vector_dot_prod(&orientation, &up) / bs_vector_length(&orientation)); 
  if(bs_vector_get_x(&orientation) > 0.0f)
  {
    angle = -angle; 
  }
  cairo_rotate(pCairo, angle); 
 
  /* face offset relaitve middle position */
  /* face off is a crap movie btw */   
  cairo_scale(pCairo, rsf, rsf); 
  cairo_translate(pCairo, 0.0, -0.30); 

  bs_blob_randomize_face_style(pBlob); 
  
  if(bs_pointmass_get_square_vel(pBlob->pMiddle) > 0.004f)
  {
    bs_blob_draw_ooh_face(pBlob, pCairo, rsf); 
  }
  else 
  {
    if(pBlob->faceStyle == BS_BLOB_FACE_STYLE_MOUTH_CLOSED)
    {
      bs_blob_draw_mouth_closed(pBlob, pCairo, rsf); 
    }
    else if(pBlob->faceStyle == BS_BLOB_FACE_STYLE_MOUTH_OPEN)
    {
      bs_blob_draw_mouth_open(pBlob, pCairo, rsf);       
    }
    else 
    {
      BS_ASSERT(FALSE);   
    }
    
    if(pBlob->eyeStyle == BS_BLOB_EYE_STYLE_BLINK)
    {    
      bs_blob_draw_eyes_blink(pBlob, pCairo, rsf); 
    }
    else if(pBlob->eyeStyle == BS_BLOB_EYE_STYLE_OPEN)
    {
      bs_blob_draw_eyes_open(pBlob, pCairo, rsf);       
    }
    else 
    {
      BS_ASSERT(FALSE);   
    }
  }
  
  cairo_restore(pCairo); 
}

static void 
bs_blob_draw_simple_body(bs_blob_t *pBlob, 
                         cairo_t *pCairo, 
                         bs_float_t scaleFactor)
{
  BS_ASSERT(pBlob != NULL);
  BS_ASSERT(pCairo != NULL);

  cairo_new_path(pCairo); 

  cairo_move_to(pCairo, 
    (bs_pointmass_get_x_pos(pBlob->pPointmasses[0]) * scaleFactor), 
    (bs_pointmass_get_y_pos(pBlob->pPointmasses[0]) * scaleFactor));  
        
  cairo_line_to(pCairo, 
    (bs_pointmass_get_x_pos(pBlob->pPointmasses[1]) * scaleFactor), 
    (bs_pointmass_get_y_pos(pBlob->pPointmasses[1]) * scaleFactor));  

  cairo_line_to(pCairo, 
    (bs_pointmass_get_x_pos(pBlob->pPointmasses[2]) * scaleFactor), 
    (bs_pointmass_get_y_pos(pBlob->pPointmasses[2]) * scaleFactor));  

  cairo_line_to(pCairo, 
    (bs_pointmass_get_x_pos(pBlob->pPointmasses[3]) * scaleFactor), 
    (bs_pointmass_get_y_pos(pBlob->pPointmasses[3]) * scaleFactor));  
  
  cairo_close_path(pCairo); 

  cairo_set_source_rgb(pCairo, 1.0, 1.0, 1.0); 
  cairo_fill_preserve(pCairo); 

  cairo_set_line_width(pCairo, 3.0);   
  cairo_set_source_rgb(pCairo, 0.0, 0.0, 0.0); 
  cairo_stroke(pCairo); 
}

#define CLAMP_INDEX(I) ((I) & 0x03)

static void 
bs_blob_draw_body(bs_blob_t *pBlob, 
                  cairo_t *pCairo, 
                  bs_float_t scaleFactor)
{
  bs_int32_t i; 
  bs_float_t px, py, nx, ny, tx, ty, cx, cy; 
  bs_pointmass_t **pPointmasses; 
  
  BS_ASSERT(pBlob != NULL);
  BS_ASSERT(pCairo != NULL);

  pPointmasses = pBlob->pPointmasses; 
  
  cairo_new_path(pCairo); 

  cairo_move_to(pCairo, 
    BS_POINTMASS_GET_X_POS(pPointmasses[0]) * scaleFactor, 
    BS_POINTMASS_GET_Y_POS(pPointmasses[0]) * scaleFactor);  
  
  for(i = 0; i < BS_BLOB_NUM_POINT_MASSES; i++)
  {
    tx = BS_POINTMASS_GET_X_POS(pPointmasses[CLAMP_INDEX(i + 1)]); 
    ty = BS_POINTMASS_GET_Y_POS(pPointmasses[CLAMP_INDEX(i + 1)]); 
    
    cx = BS_POINTMASS_GET_X_POS(pPointmasses[i]); 
    cy = BS_POINTMASS_GET_Y_POS(pPointmasses[i]); 
    
    px = (cx + tx) * 0.5f; 
    py = (cy + ty) * 0.5f;
    
    nx = cx - BS_POINTMASS_GET_X_POS(pPointmasses[CLAMP_INDEX(i + 3)]);
    nx = nx + tx - BS_POINTMASS_GET_X_POS(pPointmasses[CLAMP_INDEX(i + 2)]); 
    ny = cy - BS_POINTMASS_GET_Y_POS(pPointmasses[CLAMP_INDEX(i + 3)]); 
    ny = ny + ty - BS_POINTMASS_GET_Y_POS(pPointmasses[CLAMP_INDEX(i + 2)]); 
    
    px += nx * 0.25f; 
    py += ny * 0.25f;
    
    px *= scaleFactor; 
    py *= scaleFactor; 
    
    tx *= scaleFactor; 
    ty *= scaleFactor; 
    
    cairo_curve_to(pCairo, px, py, tx, ty, tx, ty);     
  }

  cairo_close_path(pCairo); 

  cairo_set_source_rgb(pCairo, 1.0, 1.0, 1.0); 
  cairo_fill_preserve(pCairo); 

  cairo_set_line_width(pCairo, 3.0);   
  cairo_set_source_rgb(pCairo, 0.0, 0.0, 0.0); 
  cairo_stroke(pCairo); 
}

void 
bs_blob_draw(bs_blob_t *pBlob, 
             cairo_t *pCairo, 
             bs_float_t scaleFactor)
{
  BS_ASSERT(pBlob != NULL);
  BS_ASSERT(pCairo != NULL);

  cairo_set_source_rgb(pCairo, 0.0, 0.0, 0.0); 
  
  bs_blob_draw_body(pBlob, pCairo, scaleFactor); 
  
  //bs_blob_draw_face(pBlob, pCairo, scaleFactor); 

  if (cairo_status(pCairo)) 
  {
    printf("Cairo is unhappy: %s\n",
      cairo_status_to_string (cairo_status (pCairo)));
    exit(0);
  }
}
