#include "bs_blob_collective.h"
#include "bs_mem.h"
#include "bs_pointmass.h"
#include "bs_util.h"

#include <math.h>

static bs_int32_t 
bs_blob_collective_get_unique_id()
{
  static bs_int32_t i = 0; 

  i++; 

  return i;   
}

typedef struct bs_blob_container_st
{
  bs_blob_t *pBlob;
  bs_int32_t id; 
} bs_blob_container_t; 

static bs_blob_container_t* 
bs_blob_container_create(bs_blob_t *pBlob)
{
  bs_blob_container_t *pBlobContainer; 
  
  BS_MEM_ALLOC_TYPE(pBlobContainer, bs_blob_container_t, 1); 
  pBlobContainer->pBlob = pBlob; 
  pBlobContainer->id = bs_blob_collective_get_unique_id(); 
  
  return pBlobContainer; 
}

static void 
bs_blob_container_destructor(void *pUserData)
{
  bs_blob_container_t *pBlobContainer; 
  
  BS_ASSERT(pUserData != NULL); 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  bs_blob_destroy(pBlobContainer->pBlob); 
  BS_MEM_FREE(pBlobContainer);   
}

typedef struct bs_joint_container_st
{
  bs_joint_t *pJoint; 
  bs_int32_t idA, idB; 
} bs_joint_container_t; 

static bs_joint_container_t* 
bs_joint_container_create(bs_joint_t *pJoint, 
                          bs_int32_t idA, 
                          bs_int32_t idB)
{
  bs_joint_container_t *pJointContainer; 

  BS_MEM_ALLOC_TYPE(pJointContainer, bs_joint_container_t, 1); 
  pJointContainer->pJoint = pJoint; 
  pJointContainer->idA = idA; 
  pJointContainer->idB = idB; 
  
  return pJointContainer; 
}

static void 
bs_joint_container_destroy(void *pUserData)
{
  bs_joint_container_t *pJointContainer;   
  
  BS_ASSERT(pUserData != NULL); 
  
  pJointContainer = (bs_joint_container_t*) pUserData; 
  bs_joint_destroy(pJointContainer->pJoint); 
  BS_MEM_FREE(pJointContainer); 
}

bs_blob_collective_t* 
bs_blob_collective_create(bs_float_t xPos, 
                          bs_float_t yPos, 
                          bs_list_factory_t *pFactory)
{
  bs_blob_collective_t *pCollective; 
  bs_blob_container_t *pBlobContainer; 

  BS_MEM_ALLOC_TYPE(pCollective, bs_blob_collective_t, 1); 

  pCollective->pFactory = pFactory; 
  
  pCollective->pBlobs = bs_list_factory_get_list(pFactory); 
  
  pBlobContainer = bs_blob_container_create(bs_blob_create(xPos, yPos, 0.4f)); 
  bs_list_add_element_to_front(pCollective->pBlobs, (void*)pBlobContainer, 
    bs_blob_container_destructor); 
  
  pCollective->pJoints = bs_list_factory_get_list(pFactory); 
  
  return pCollective; 
}

void 
bs_blob_collective_destroy(bs_blob_collective_t *pCollective)
{
  BS_ASSERT(pCollective != NULL);
  BS_ASSERT(pCollective->pBlobs != NULL); 
  
  bs_list_factory_give_list(pCollective->pFactory, pCollective->pBlobs); 
}

typedef struct bs_blob_collective_find_largest_st
{
  bs_float_t maxRadius; 
  bs_blob_t *pMotherBlob; 
  bs_int32_t id; 
} bs_blob_collective_find_largest_t; 

static void 
bs_blob_collective_find_largest_cb(void *pUserData, 
                                   void *pAccumulator)
{
  bs_blob_container_t *pBlobContainer; 
  bs_blob_collective_find_largest_t *pFindData;   
  bs_float_t radius; 
  
  BS_ASSERT(pUserData != NULL); 
  BS_ASSERT(pAccumulator != NULL); 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  pFindData = (bs_blob_collective_find_largest_t*) pAccumulator; 
  
  radius = bs_blob_get_radius(pBlobContainer->pBlob); 
  if(radius > pFindData->maxRadius)
  {
    pFindData->maxRadius = radius; 
    pFindData->pMotherBlob = pBlobContainer->pBlob;    
    pFindData->id = pBlobContainer->id;     
  }
}

typedef struct bs_blob_collective_create_new_joints_st
{
  bs_blob_collective_t *pCollective; 
  bs_blob_container_t *pNewBlobContainer; 
} bs_blob_collective_create_new_joints_t; 

static void 
bs_blob_collective_create_new_joints_cb(void *pUserData, 
                                        void *pAccumulator)
{
  bs_blob_container_t *pListBlob; 
  bs_blob_collective_create_new_joints_t *pCreateData; 
  bs_pointmass_t *pListBlobPointmass, *pNewBlobPointmass; 
  bs_joint_t *pNewJoint; 
  bs_joint_container_t *pJointContainer; 
  bs_float_t dist; 
  
  BS_ASSERT(pUserData != NULL); 
  BS_ASSERT(pAccumulator != NULL); 
  
  pListBlob = (bs_blob_container_t*) pUserData; 
  pCreateData = (bs_blob_collective_create_new_joints_t*) pAccumulator; 
  
  pListBlobPointmass = bs_blob_get_middle(pListBlob->pBlob); 
  pNewBlobPointmass = bs_blob_get_middle(pCreateData->pNewBlobContainer->pBlob); 
 
  pNewJoint = bs_joint_create(pListBlobPointmass, pNewBlobPointmass, 
    BS_JOINT_NO_CONST, BS_JOINT_NO_CONST); 
  dist = bs_blob_get_radius(pListBlob->pBlob) + 
    bs_blob_get_radius(pCreateData->pNewBlobContainer->pBlob); 
  bs_joint_set_consts(pNewJoint, dist, BS_JOINT_NO_CONST); 
  
  pJointContainer = bs_joint_container_create(pNewJoint, 
    pCreateData->pNewBlobContainer->id, pListBlob->id);
    
  bs_list_add_element_to_front(pCreateData->pCollective->pJoints, 
    (void*) pJointContainer, bs_joint_container_destroy);   
}

static void 
bs_blob_collective_add_new_blob(bs_blob_collective_t *pCollective, 
                                bs_float_t xPos, 
                                bs_float_t yPos, 
                                bs_float_t radius)
{
  bs_blob_t *pNewBlob; 
  bs_blob_container_t *pNewBlobContainer; 
  bs_blob_collective_create_new_joints_t createData; 
  
  pNewBlob = bs_blob_create(xPos, yPos, radius); 
  pNewBlobContainer = bs_blob_container_create(pNewBlob); 

  createData.pCollective = pCollective; 
  createData.pNewBlobContainer = pNewBlobContainer; 
  bs_list_map_acc(pCollective->pBlobs, bs_blob_collective_create_new_joints_cb, 
    &createData); 
  
  bs_list_add_element_to_front(pCollective->pBlobs, pNewBlobContainer, 
    bs_blob_container_destructor); 
}

static bs_bool_t
bs_blob_collective_remove_joints_cb(void *pUserData, 
                                    bs_int32_t blobId)
{
  bs_joint_container_t *pJointContainer; 
  
  pJointContainer = (bs_joint_container_t*) pUserData; 
  BS_ASSERT(pJointContainer != NULL); 
  
  if(pJointContainer->idA == blobId || 
     pJointContainer->idB == blobId)
  {
    return TRUE; 
  }    
  return FALSE; 
}

static bs_bool_t 
bs_blob_collective_remove_blob_cb(void *pUserData, 
                                  bs_int32_t blobId, 
                                  bs_bool_t *pCont)
{
  bs_blob_container_t *pBlobContainer; 

  pBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pBlobContainer != NULL); 
  
  if(pBlobContainer->id == blobId)
  {
    *pCont = FALSE; 
    return TRUE; 
  }
  
  *pCont = TRUE; 
  return FALSE; 
}

static bs_float_t 
bs_blob_collective_split_get_radius(bs_float_t motherRadius, 
                                    bs_float_t partOfMother)
{
  bs_float_t motherVolume; 
  
  motherVolume = motherRadius * motherRadius * M_PI; 
  return sqrt(motherVolume * partOfMother / M_PI); 
}

void 
bs_blob_collective_split(bs_blob_collective_t *pCollective)
{
  bs_blob_collective_find_largest_t findData; 
  bs_float_t motherXPos, motherYPos;
  bs_float_t motherRadius; 
  bs_float_t aRadius, bRadius, largestRadius; 
  
  BS_ASSERT(pCollective != NULL);

  findData.maxRadius = -1.0f; 
  findData.pMotherBlob = NULL; 
  
  bs_list_map_acc(pCollective->pBlobs, bs_blob_collective_find_largest_cb, &findData); 
  
  BS_ASSERT(findData.pMotherBlob != NULL); 
  
  motherXPos = bs_blob_get_x_pos(findData.pMotherBlob); 
  motherYPos = bs_blob_get_y_pos(findData.pMotherBlob); 
  motherRadius = bs_blob_get_radius(findData.pMotherBlob); 
  
  aRadius = 5000.0f + 1000.0f - (bs_float_t)(random() % 2000); 
  aRadius /= 10000.0f;
  bRadius = 1.0f - aRadius; 
  
  largestRadius = BS_UTIL_MAX(aRadius, bRadius); 
  
  bs_blob_collective_add_new_blob(pCollective, motherXPos + largestRadius * 0.35f, motherYPos, 
    bs_blob_collective_split_get_radius(motherRadius, aRadius));
  bs_blob_collective_add_new_blob(pCollective, motherXPos - largestRadius * 0.35f, motherYPos, 
    bs_blob_collective_split_get_radius(motherRadius, bRadius)); 

  bs_list_unlink_i_cont(pCollective->pBlobs, bs_blob_collective_remove_blob_cb, findData.id); 
  
  printf("%d\n", bs_list_get_length(pCollective->pBlobs)); 
}

typedef struct bs_blob_collective_join_find_smallest_st
{
  bs_blob_container_t *pSmallest; 
  bs_float_t minRadius; 
} bs_blob_collective_join_find_smallest_t; 

static void 
bs_blob_collective_join_find_smallest_cb(void *pUserData, 
                                         void *pAccumulator)
{
  bs_blob_container_t *pListBlobContainer; 
  bs_blob_collective_join_find_smallest_t *pFindData; 
  bs_float_t radius; 
  
  pListBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pListBlobContainer != NULL);
  pFindData = (bs_blob_collective_join_find_smallest_t*) pAccumulator; 
  BS_ASSERT(pFindData != NULL);

  radius = bs_blob_get_radius(pListBlobContainer->pBlob); 
  if(radius < pFindData->minRadius)
  {
    pFindData->minRadius = radius; 
    pFindData->pSmallest = pListBlobContainer; 
  }    
}

typedef struct bs_blob_collective_join_find_closest_st
{
  bs_blob_container_t *pClosest; 
  bs_float_t minDist; 
  bs_pointmass_t *pBlobMiddle; 
  bs_int32_t smallestId; 
} bs_blob_collective_join_find_closest_t; 

static void 
bs_blob_collective_join_find_closest_cb(void *pUserData, 
                                        void *pAccumulator)
{
  bs_blob_container_t *pListBlobContainer; 
  bs_blob_collective_join_find_closest_t *pFindData; 
  bs_pointmass_t *pListPointmass; 
  bs_float_t aXbX, aYbY, dist; 
  
  pListBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pListBlobContainer != NULL);
  pFindData = (bs_blob_collective_join_find_closest_t*) pAccumulator; 
  BS_ASSERT(pFindData != NULL);
  
  if(pListBlobContainer->id == pFindData->smallestId)
  {
    return; 
  }
  
  pListPointmass = bs_blob_get_middle(pListBlobContainer->pBlob); 
  
  aXbX = bs_pointmass_get_x_pos(pListPointmass) - bs_pointmass_get_x_pos(pFindData->pBlobMiddle); 
  aYbY = bs_pointmass_get_y_pos(pListPointmass) - bs_pointmass_get_y_pos(pFindData->pBlobMiddle); 
  
  dist = aXbX * aXbX + aYbY * aYbY; 
  
  if(dist < pFindData->minDist)
  {
    pFindData->minDist = dist; 
    pFindData->pClosest = pListBlobContainer; 
  }    
}

static bs_float_t
bs_blob_collective_join_get_radius(bs_float_t smallestRadius, 
                                   bs_float_t closestRadius)
{
  bs_float_t newRadius; 
  bs_float_t smallestVolume, closestVolume; 
  
  smallestVolume = smallestRadius * smallestRadius * M_PI; 
  closestVolume = closestRadius * closestRadius * M_PI; 
  newRadius = sqrt((smallestVolume + closestVolume) / M_PI); 
  
  return newRadius; 
}

static void 
bs_blob_collective_join_blobs(bs_blob_collective_t *pCollective, 
                              bs_blob_container_t *pBlobContainerA, 
                              bs_blob_container_t *pBlobContainerB)
{
  bs_float_t newRadius, newXPos, newYPos; 
  
  BS_ASSERT(pCollective != NULL);
  BS_ASSERT(pBlobContainerA != NULL);   
  BS_ASSERT(pBlobContainerB != NULL);   
  
  newXPos = bs_blob_get_x_pos(pBlobContainerA->pBlob); 
  newYPos = bs_blob_get_y_pos(pBlobContainerA->pBlob); 

  newRadius = bs_blob_collective_join_get_radius(
    bs_blob_get_radius(pBlobContainerA->pBlob), 
    bs_blob_get_radius(pBlobContainerB->pBlob)); 
   
  bs_list_unlink_i(pCollective->pJoints, bs_blob_collective_remove_joints_cb, pBlobContainerA->id);
  bs_list_unlink_i_cont(pCollective->pBlobs, bs_blob_collective_remove_blob_cb, pBlobContainerA->id); 
  
  bs_list_unlink_i(pCollective->pJoints, bs_blob_collective_remove_joints_cb, pBlobContainerB->id);
  bs_list_unlink_i_cont(pCollective->pBlobs, bs_blob_collective_remove_blob_cb, pBlobContainerB->id); 

  bs_blob_collective_add_new_blob(pCollective, newXPos, newYPos, newRadius); 
}

void 
bs_blob_collective_join(bs_blob_collective_t *pCollective)
{
  bs_blob_collective_join_find_smallest_t findSmallest; 
  bs_blob_collective_join_find_closest_t findClosest; 
  
  BS_ASSERT(pCollective != NULL);  
  BS_ASSERT(bs_list_get_length(pCollective->pBlobs) > 0);   

  if(bs_list_get_length(pCollective->pBlobs) == 1)
  {
    return; 
  }
  
  findSmallest.pSmallest = NULL; 
  findSmallest.minRadius = 10000.0f; 
  bs_list_map_acc(pCollective->pBlobs, bs_blob_collective_join_find_smallest_cb, &findSmallest); 
  BS_ASSERT(findSmallest.pSmallest != NULL); 
  
  findClosest.pClosest = NULL; 
  findClosest.minDist = 10000.0f; 
  findClosest.pBlobMiddle = bs_blob_get_middle(findSmallest.pSmallest->pBlob); 
  findClosest.smallestId = findSmallest.pSmallest->id;
  bs_list_map_acc(pCollective->pBlobs, bs_blob_collective_join_find_closest_cb, &findClosest); 
  BS_ASSERT(findClosest.pClosest != NULL); 
  
  bs_blob_collective_join_blobs(pCollective, findClosest.pClosest, findSmallest.pSmallest); 
}

static bs_bool_t
bs_blob_collective_merge_test2(void *pUserData, 
                               void *pDataArgA, 
                               void *pDataArgB)
{
  bs_blob_container_t *pBlobContainerA, *pBlobContainerB; 
  bs_blob_collective_t *pCollective; 
  bs_pointmass_t *pMiddleA, *pMiddleB; 
  bs_float_t minRadius; 
  
  BS_ASSERT(pUserData != NULL);
  pBlobContainerA = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pDataArgA != NULL); 
  pBlobContainerB = (bs_blob_container_t*) pDataArgB; 
  BS_ASSERT(pDataArgB != NULL); 
  pCollective = (bs_blob_collective_t*) pDataArgA; 
  
  if(pBlobContainerA->id == pBlobContainerB->id)
  {
    return TRUE; 
  }
  
  pMiddleA = bs_blob_get_middle(pBlobContainerA->pBlob); 
  pMiddleB = bs_blob_get_middle(pBlobContainerB->pBlob); 
  
  minRadius = BS_UTIL_MIN(bs_blob_get_radius(pBlobContainerA->pBlob), 
    bs_blob_get_radius(pBlobContainerB->pBlob)); 
  
  if(bs_pointmass_get_dist(pMiddleA, pMiddleB) < minRadius * 0.5)
  {
    bs_blob_collective_join_blobs(pCollective, pBlobContainerA, pBlobContainerB); 
    return FALSE; 
  }
  
  return TRUE; 
}

static bs_bool_t 
bs_blob_collective_merge_test(void *pUserData, 
                              void *pDataArg)
{
  bs_blob_container_t *pBlobContainer; 
  bs_blob_collective_t *pCollective; 
  bs_int32_t listLength; 
  
  BS_ASSERT(pUserData != NULL); 
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pDataArg != NULL);
  pCollective = (bs_blob_collective_t*) pDataArg;   
  
  listLength = bs_list_get_length(pCollective->pBlobs); 
  
  bs_list_map_pp_cont(pCollective->pBlobs, bs_blob_collective_merge_test2, 
    (void*)pCollective, (void*)pBlobContainer); 
  
  if(listLength > bs_list_get_length(pCollective->pBlobs))
  {
    return FALSE; 
  }
  
  return TRUE; 
}

void 
bs_blob_collective_merge_test_blobs(bs_blob_collective_t *pCollective)
{
  bs_int32_t listLength; 
  
  BS_ASSERT(pCollective != NULL); 
  
  return;
  
  do 
  {
    listLength = bs_list_get_length(pCollective->pBlobs); 
    bs_list_map_p_cont(pCollective->pBlobs, bs_blob_collective_merge_test, (void*)pCollective);   
  } while(listLength > bs_list_get_length(pCollective->pBlobs));
}

static void 
bs_blob_collective_move_cb(void *pUserData, 
                           bs_float_t dt)
{
  bs_blob_container_t *pBlobContainer; 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pBlobContainer != NULL); 

  bs_blob_move(pBlobContainer->pBlob, dt); 
}

void 
bs_blob_collective_move(bs_blob_collective_t *pCollective, 
                        bs_float_t dt)
{
  BS_ASSERT(pCollective != NULL); 
  bs_list_map_f(pCollective->pBlobs, bs_blob_collective_move_cb, dt);
}

static void 
bs_blob_colelctive_blob_sc_cb(void *pUserData, 
                              void *pDataArg)
{
  bs_blob_container_t *pBlobContainer; 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pBlobContainer != NULL); 
  
  bs_blob_sc(pBlobContainer->pBlob, (bs_env_t*) pDataArg); 
}

static void 
bs_blob_collective_joint_sc_cb(void *pUserData)
{
  bs_joint_container_t *pJointContainer; 

  pJointContainer = (bs_joint_container_t*) pUserData;
  BS_ASSERT(pJointContainer != NULL);

  bs_joint_sc(pJointContainer->pJoint);   
}

static void 
bs_blob_collective_collision_blob_blob_cb(void *pUserData, 
                                          void *pDataArg)
{
  bs_blob_container_t *pTested, *pAgainst; 
  
  BS_ASSERT(pUserData != NULL);
  BS_ASSERT(pDataArg != NULL);   

  pAgainst = (bs_blob_container_t*) pUserData; 
  pTested = (bs_blob_container_t*) pDataArg;   
  
  if(pAgainst->id == pTested->id)
  {
    return; 
  }
  
  bs_blob_collision_test(pAgainst->pBlob, pTested->pBlob); 
}

static void 
bs_blob_collective_collision_list_blob_cb(void *pUserData, 
                                          void *pDataArg)
{
  bs_blob_container_t *pBlobContainer; 
  bs_blob_collective_t *pCollective; 
  
  BS_ASSERT(pUserData != NULL);
  BS_ASSERT(pDataArg != NULL);   
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  pCollective = (bs_blob_collective_t*) pDataArg; 
  
  bs_list_map_p(pCollective->pBlobs, bs_blob_collective_collision_blob_blob_cb, 
    pBlobContainer); 
}

void 
bs_blob_collective_sc(bs_blob_collective_t *pCollective, 
                      bs_env_t *pEnv)
{
  BS_ASSERT(pCollective != NULL); 

  bs_list_map_p(pCollective->pBlobs, bs_blob_colelctive_blob_sc_cb, (void*) pEnv); 
  
  bs_list_map_p(pCollective->pBlobs, bs_blob_collective_collision_list_blob_cb, 
    (void*) pCollective);     
}

static void 
bs_blob_collective_set_force_cb(void *pUserData, 
                                void *pDataArg)
{
  bs_blob_container_t *pBlobContainer; 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pBlobContainer != NULL); 
  
  bs_blob_set_force(pBlobContainer->pBlob, (bs_vector_t*) pDataArg); 
}

void 
bs_blob_collective_set_force(bs_blob_collective_t *pCollective, 
                             bs_vector_t *pForce)
{
  BS_ASSERT(pCollective != NULL); 
  bs_list_map_p(pCollective->pBlobs, bs_blob_collective_set_force_cb, (void*) pForce); 
}

static void 
bs_blob_collective_add_force_cb(void *pUserData, 
                                void *pDataArg)
{
  bs_blob_container_t *pBlobContainer; 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pBlobContainer != NULL); 
  
  bs_blob_add_force(pBlobContainer->pBlob, (bs_vector_t*) pDataArg); 
}

void 
bs_blob_collective_add_force(bs_blob_collective_t *pCollective, 
                             bs_vector_t *pForce)
{
  BS_ASSERT(pCollective != NULL); 
  bs_list_map_p(pCollective->pBlobs, bs_blob_collective_add_force_cb, (void*) pForce); 
}

static void 
bs_blob_collective_draw_blob_cb(void *pUserData, 
                                bs_float_t scaleFactor, 
                                void *pDataArg)
{
  bs_blob_container_t *pBlobContainer; 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pBlobContainer != NULL); 
  
  bs_blob_draw(pBlobContainer->pBlob, (cairo_t*) pDataArg, scaleFactor); 
}

static void 
bs_blob_collective_draw_joint_cb(void *pUserData, 
                                 bs_float_t scaleFactor, 
                                 void *pDataArg)
{
  bs_joint_container_t *pJointContainer; 
  
  pJointContainer = (bs_joint_container_t*) pUserData; 
  BS_ASSERT(pJointContainer != NULL);

  bs_joint_draw(pJointContainer->pJoint, (cairo_t*) pDataArg, scaleFactor);   
}

void 
bs_blob_collective_draw(bs_blob_collective_t *pCollective, 
                        cairo_t *pCairo, 
                        bs_float_t scaleFactor)
{
  bs_list_map_fp(pCollective->pBlobs, bs_blob_collective_draw_blob_cb, 
    scaleFactor, (void*) pCairo);  
}

static void 
bs_blob_collective_larger_blobs_cb(void *pUserData)
{
  bs_blob_container_t *pContainer; 
  
  pContainer = (bs_blob_container_t*) pUserData; 
  bs_blob_scale(pContainer->pBlob, 1.1f); 
}

static void 
bs_blob_collective_smaller_blobs_cb(void *pUserData)
{
  bs_blob_container_t *pContainer; 
  
  pContainer = (bs_blob_container_t*) pUserData; 
  bs_blob_scale(pContainer->pBlob, 0.9f); 
}

void 
bs_blob_collective_larger_blobs(bs_blob_collective_t *pCollective)
{
  bs_list_map(pCollective->pBlobs, bs_blob_collective_larger_blobs_cb); 
}

void 
bs_blob_collective_smaller_blobs(bs_blob_collective_t *pCollective)
{
  bs_list_map(pCollective->pBlobs, bs_blob_collective_smaller_blobs_cb);   
}
